#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import logging
from typing import Optional, cast

import botocore.exceptions
import click

from aws_orbit import bundle, remote
from aws_orbit.messages import MessagesContext
from aws_orbit.models.context import Context, ContextSerDe, FoundationContext
from aws_orbit.services import cfn, codebuild, ecr, elb, s3, secretsmanager, ssm

_logger: logging.Logger = logging.getLogger(__name__)


def destroy_toolkit(
    env_name: str,
    top_level: str = "orbit",
    toolkit_bucket: Optional[str] = None,
    cdk_toolkit_bucket: Optional[str] = None,
) -> None:
    try:
        if toolkit_bucket:
            s3.delete_bucket(bucket=toolkit_bucket)
    except Exception as ex:
        _logger.debug("Skipping Toolkit bucket deletion. Cause: %s", ex)
    try:
        if cdk_toolkit_bucket:
            s3.delete_bucket(bucket=cdk_toolkit_bucket)
    except Exception as ex:
        _logger.debug("Skipping CDK Toolkit bucket deletion. Cause: %s", ex)
    toolkit_stack_name: str = f"{top_level}-{env_name}-toolkit"
    if cfn.does_stack_exist(stack_name=toolkit_stack_name):
        cfn.destroy_stack(stack_name=toolkit_stack_name)
    ssm.cleanup_env(env_name=env_name, top_level=top_level)


def destroy_remaining_resources(env_name: str, top_level: str = "orbit") -> None:
    ecr.cleanup_remaining_repos(env_name=env_name)
    env_cdk_toolkit: str = f"{top_level}-{env_name}-cdk-toolkit"
    if cfn.does_stack_exist(stack_name=env_cdk_toolkit):
        cfn.destroy_stack(stack_name=env_cdk_toolkit)
    destroy_toolkit(env_name=env_name)


def destroy_teams(env: str, debug: bool) -> None:
    with MessagesContext("Destroying", debug=debug) as msg_ctx:
        ssm.cleanup_changeset(env_name=env)

        if not ssm.list_parameters(prefix=f"/orbit/{env}/teams/"):
            msg_ctx.info(f"No {env} Teams found.")
            msg_ctx.progress(100)
            return

        context: "Context" = ContextSerDe.load_context_from_ssm(env_name=env, type=Context)
        msg_ctx.info("Context loaded")
        msg_ctx.info(f"Teams: {','.join([t.name for t in context.teams])}")
        msg_ctx.progress(2)

        msg_ctx.progress(4)

        if any(cfn.does_stack_exist(stack_name=t.stack_name) for t in context.teams):
            bundle_path = bundle.generate_bundle(command_name="destroy", context=context)
            msg_ctx.progress(5)

            buildspec = codebuild.generate_spec(
                context=context,
                plugins=True,
                cmds_build=[f"orbit remote --command destroy_teams {env}"],
                changeset=None,
            )
            remote.run(
                command_name="destroy",
                context=context,
                bundle_path=bundle_path,
                buildspec=buildspec,
                codebuild_log_callback=msg_ctx.progress_bar_callback,
                timeout=120,
            )
        msg_ctx.progress(95)

        msg_ctx.info("Teams Destroyed")
        msg_ctx.progress(100)


def destroy_env(env: str, preserve_credentials: bool, debug: bool) -> None:
    with MessagesContext("Destroying", debug=debug) as msg_ctx:
        ssm.cleanup_changeset(env_name=env)
        ssm.cleanup_manifest(env_name=env)

        if ssm.does_parameter_exist(name=f"/orbit/{env}/context") is False:
            msg_ctx.info(f"Environment {env} not found. Destroying only possible remaining resources.")
            elb.delete_load_balancers(env_name=env)
            destroy_remaining_resources(env_name=env, top_level="orbit")
            msg_ctx.progress(100)
            return

        context: "Context" = ContextSerDe.load_context_from_ssm(env_name=env, type=Context)
        msg_ctx.info("Context loaded")
        msg_ctx.info(f"Teams: {','.join([t.name for t in context.teams])}")
        msg_ctx.progress(2)

        if any(cfn.does_stack_exist(stack_name=t.stack_name) for t in context.teams):
            raise click.ClickException("Found Teams dependent on the Envrionment.")

        if (
            cfn.does_stack_exist(stack_name=context.env_stack_name)
            or cfn.does_stack_exist(stack_name=context.toolkit.stack_name)
            or cfn.does_stack_exist(stack_name=context.cdk_toolkit.stack_name)
        ):
            bundle_path = bundle.generate_bundle(command_name="destroy", context=context)
            msg_ctx.progress(5)

            buildspec = codebuild.generate_spec(
                context=context,
                plugins=True,
                cmds_build=[f"orbit remote --command destroy_env {env}"],
                changeset=None,
            )
            remote.run(
                command_name="destroy",
                context=context,
                bundle_path=bundle_path,
                buildspec=buildspec,
                codebuild_log_callback=msg_ctx.progress_bar_callback,
                timeout=120,
            )

        if not preserve_credentials:
            secretsmanager.delete_docker_credentials(secret_id=f"orbit-{context.name}-docker-credentials")
            _logger.info("Removed docker credentials from SecretsManager")

        msg_ctx.info("Env destroyed")
        msg_ctx.progress(95)

        try:
            destroy_toolkit(
                env_name=context.name,
                toolkit_bucket=context.toolkit.s3_bucket,
                cdk_toolkit_bucket=context.cdk_toolkit.s3_bucket,
            )
        except botocore.exceptions.ClientError as ex:
            error = ex.response["Error"]
            if "does not exist" not in error["Message"]:
                raise
            _logger.debug(f"Skipping toolkit destroy: {error['Message']}")
        msg_ctx.info("Toolkit destroyed")
        ssm.cleanup_env(env_name=context.name)

        msg_ctx.progress(100)


def destroy_foundation(env: str, debug: bool) -> None:
    with MessagesContext("Destroying", debug=debug) as msg_ctx:
        ssm.cleanup_changeset(env_name=env, top_level="orbit-f")
        ssm.cleanup_manifest(env_name=env, top_level="orbit-f")

        if ssm.does_parameter_exist(name=f"/orbit-f/{env}/context") is False:
            msg_ctx.info(f"Foundation {env} not found. Destroying only possible remaining resources.")
            destroy_remaining_resources(env_name=env, top_level="orbit-f")
            msg_ctx.progress(100)
            return

        context: "FoundationContext" = ContextSerDe.load_context_from_ssm(env_name=env, type=FoundationContext)
        msg_ctx.info("Context loaded")
        msg_ctx.progress(2)

        msg_ctx.progress(4)

        if (
            cfn.does_stack_exist(stack_name=cast(str, context.stack_name))
            or cfn.does_stack_exist(stack_name=context.toolkit.stack_name)
            or cfn.does_stack_exist(stack_name=context.cdk_toolkit.stack_name)
        ):
            bundle_path = bundle.generate_bundle(command_name="destroy", context=cast(Context, context))
            msg_ctx.progress(5)

            buildspec = codebuild.generate_spec(
                context=cast(Context, context),
                plugins=False,
                cmds_build=[f"orbit remote --command destroy_foundation {env}"],
                changeset=None,
            )
            remote.run(
                command_name="destroy",
                context=cast(Context, context),
                bundle_path=bundle_path,
                buildspec=buildspec,
                codebuild_log_callback=msg_ctx.progress_bar_callback,
                timeout=45,
            )
        msg_ctx.info("Foundation destroyed")
        msg_ctx.progress(95)

        try:
            destroy_toolkit(
                env_name=context.name,
                top_level="orbit-f",
                toolkit_bucket=context.toolkit.s3_bucket,
                cdk_toolkit_bucket=context.cdk_toolkit.s3_bucket,
            )
        except botocore.exceptions.ClientError as ex:
            error = ex.response["Error"]
            if "does not exist" not in error["Message"]:
                raise
            _logger.debug(f"Skipping toolkit destroy: {error['Message']}")
        msg_ctx.info("Toolkit destroyed")
        ssm.cleanup_env(env_name=context.name, top_level="orbit-f")

        msg_ctx.progress(100)


def destroy_credentials(env: str, registry: str, debug: bool) -> None:
    with MessagesContext("Destroying", debug=debug) as msg_ctx:
        context: "Context" = ContextSerDe.load_context_from_ssm(env_name=env, type=Context)
        msg_ctx.info("Context loaded")
        msg_ctx.progress(2)

        msg_ctx.progress(4)

        if any(cfn.does_stack_exist(stack_name=t.stack_name) for t in context.teams):
            bundle_path = bundle.generate_bundle(command_name="destroy", context=context)
            msg_ctx.progress(5)

            buildspec = codebuild.generate_spec(
                context=context,
                plugins=True,
                cmds_build=[f"orbit remote --command destroy_credentials {env} {registry}"],
                changeset=None,
            )
            remote.run(
                command_name="destroy",
                context=context,
                bundle_path=bundle_path,
                buildspec=buildspec,
                codebuild_log_callback=msg_ctx.progress_bar_callback,
                timeout=10,
            )
        msg_ctx.progress(95)

        msg_ctx.info("Registry Credentials Destroyed")
        msg_ctx.progress(100)



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import logging
import os
from typing import cast

from aws_orbit import ORBIT_CLI_ROOT, cdk, cleanup
from aws_orbit.models.context import FoundationContext
from aws_orbit.remote_files.cert import check_cert
from aws_orbit.services import cfn, ssm, vpc

_logger: logging.Logger = logging.getLogger(__name__)


def _fetch_vpc_id(context: "FoundationContext") -> str:
    return cast(str, ssm.get_parameter(name=cast(str, context.resources_ssm_parameter_name))["VpcId"])


def deploy(context: "FoundationContext") -> None:
    stack_name: str = cast(str, context.stack_name)

    _logger.debug("Deploying self signed cert...")
    ssl_cert_arn = check_cert(context=context)

    _logger.debug("Deploying %s Foundation...", stack_name)
    cdk.deploy(
        context=context,
        stack_name=stack_name,
        app_filename=os.path.join(ORBIT_CLI_ROOT, "remote_files", "cdk", "foundation.py"),
        args=[context.name, ssl_cert_arn],
    )
    _logger.debug("Enabling private dns for codeartifact vpc endpoints")
    vpc_id: str = _fetch_vpc_id(context=context)
    vpc.modify_vpc_endpoint(vpc_id=vpc_id, service_name="codeartifact.repositories", private_dns_enabled=True)
    vpc.modify_vpc_endpoint(vpc_id=vpc_id, service_name="codeartifact.api", private_dns_enabled=True)


def destroy(context: "FoundationContext") -> None:
    if cfn.does_stack_exist(stack_name=cast(str, context.stack_name)):
        cleanup.foundation_remaining_dependencies(context=context)
        cleanup.delete_cert_from_iam(context=context)
        _logger.debug("Destroying Foundation...")
        cdk.destroy(
            context=context,
            stack_name=cast(str, context.stack_name),
            app_filename=os.path.join(ORBIT_CLI_ROOT, "remote_files", "cdk", "foundation.py"),
            args=[context.name],
        )



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import logging
from typing import Optional, cast

import aws_cdk.aws_kms as kms
import aws_cdk.aws_s3 as s3
import aws_cdk.core as core

_logger: logging.Logger = logging.getLogger(__name__)


class S3Builder:
    @staticmethod
    def build_s3_bucket(
        scope: core.Construct, id: str, name: str, scratch_retention_days: int, kms_key: kms.Key
    ) -> s3.Bucket:
        _logger.debug(f"Creating scratch bucket {name}")

        return s3.Bucket(
            scope=scope,
            id=id,
            bucket_name=name,
            access_control=s3.BucketAccessControl.BUCKET_OWNER_FULL_CONTROL,
            block_public_access=s3.BlockPublicAccess.BLOCK_ALL,
            removal_policy=core.RemovalPolicy.RETAIN,
            lifecycle_rules=[s3.LifecycleRule(expiration=core.Duration.days(scratch_retention_days))],
            encryption=s3.BucketEncryption.KMS,
            encryption_key=cast(
                Optional[kms.IKey],
                kms_key,
            ),
        )



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import logging
from datetime import datetime, timedelta, timezone
from typing import Any, Callable, Dict, List, Optional

from aws_orbit.models.context import Context
from aws_orbit.services import cloudwatch, codebuild, s3

_logger: logging.Logger = logging.getLogger(__name__)


def _print_codebuild_logs(
    events: List[cloudwatch.CloudWatchEvent],
    codebuild_log_callback: Callable[[str], None],
) -> None:
    for event in events:
        msg = event.message[:-1] if event.message.endswith("\n") else event.message
        _logger.debug("[CODEBUILD] %s", msg)
        codebuild_log_callback(msg)


def _wait_execution(
    build_id: str,
    stream_name_prefix: str,
    codebuild_log_callback: Optional[Callable[[str], None]] = None,
) -> None:
    start_time: Optional[datetime] = None
    stream_name: Optional[str] = None
    for status in codebuild.wait(build_id=build_id):
        if codebuild_log_callback is not None and status.logs.enabled and status.logs.group_name:
            if stream_name is None:
                stream_name = cloudwatch.get_stream_name_by_prefix(
                    group_name=status.logs.group_name,
                    prefix=f"{stream_name_prefix}/",
                )
            if stream_name is not None:
                events = cloudwatch.get_log_events(
                    group_name=status.logs.group_name,
                    stream_name=stream_name,
                    start_time=start_time,
                )
                _print_codebuild_logs(events=events.events, codebuild_log_callback=codebuild_log_callback)
                if events.last_timestamp is not None:
                    start_time = events.last_timestamp + timedelta(milliseconds=1)


def _execute_codebuild(
    context: "Context",
    command_name: str,
    buildspec: codebuild.SPEC_TYPE,
    timeout: int,
    overrides: Optional[Dict[str, Any]] = None,
    codebuild_log_callback: Optional[Callable[[str], None]] = None,
) -> None:
    if context.toolkit.s3_bucket is None:
        raise ValueError(f"context.toolkit.s3_bucket: {context.toolkit.s3_bucket}")
    bundle_location = f"{context.toolkit.s3_bucket}/cli/remote/{command_name}/bundle.zip"
    _logger.debug("bundle_location: %s", bundle_location)
    stream_name_prefix = f"{command_name}-{int(datetime.now(timezone.utc).timestamp() * 1_000_000)}"
    _logger.debug("stream_name_prefix: %s", stream_name_prefix)
    build_id = codebuild.start(
        context=context,
        project_name=context.toolkit.codebuild_project,
        stream_name=stream_name_prefix,
        bundle_location=bundle_location,
        buildspec=buildspec,
        timeout=timeout,
        overrides=overrides,
    )
    _wait_execution(
        build_id=build_id,
        stream_name_prefix=stream_name_prefix,
        codebuild_log_callback=codebuild_log_callback,
    )


def run(
    command_name: str,
    context: "Context",
    bundle_path: str,
    buildspec: codebuild.SPEC_TYPE,
    timeout: int,
    overrides: Optional[Dict[str, Any]] = None,
    codebuild_log_callback: Optional[Callable[[str], None]] = None,
) -> None:
    if context.toolkit.s3_bucket is None:
        raise ValueError(f"context.toolkit.s3_bucket: {context.toolkit.s3_bucket}")
    bucket: str = context.toolkit.s3_bucket
    key: str = f"cli/remote/{command_name}/bundle.zip"
    s3.delete_objects(bucket=bucket, keys=[key])
    s3.upload_file(src=bundle_path, bucket=bucket, key=key)
    _execute_codebuild(
        context=context,
        command_name=command_name,
        buildspec=buildspec,
        codebuild_log_callback=codebuild_log_callback,
        timeout=timeout,
        overrides=overrides,
    )
    s3.delete_objects(bucket=bucket, keys=[key])



import json
import logging
import os
import subprocess
import time
from typing import Any, Dict, List, Optional, cast

import boto3
from kubernetes import client, config, dynamic
from kubernetes.client import api_client
from kubernetes.client.rest import ApiException

logger = logging.getLogger()
logger.setLevel(logging.INFO)

ORBIT_ENV = os.environ.get("ORBIT_ENV")
ACCOUNT_ID = os.environ.get("ACCOUNT_ID")
REGION = os.environ.get("REGION")
ROLE_PREFIX = os.environ.get("ROLE_PREFIX")
ORBIT_API_VERSION = os.environ.get("ORBIT_API_VERSION", "v1")
ORBIT_API_GROUP = os.environ.get("ORBIT_API_GROUP", "orbit.aws")
ORBIT_SYSTEM_NAMESPACE = os.environ.get("ORBIT_SYSTEM_NAMESPACE", "orbit-system")
ORBIT_STATE_PATH = os.environ.get("ORBIT_STATE_PATH", "/state")
USERSPACE_CR_KIND = "UserSpace"
KUBECONFIG_PATH = "/tmp/.kubeconfig"


ssm = boto3.client("ssm")
context = ssm.get_parameter(Name=f"/orbit/{ORBIT_ENV}/context")
EFS_FS_ID = json.loads(context.get("Parameter").get("Value")).get("SharedEfsFsId")


def handler(event: Dict[str, Any], context: Optional[Dict[str, Any]]) -> Any:
    create_kubeconfig()

    api_CoreV1 = client.CoreV1Api()
    userspace_dc = dynamic.DynamicClient(client=api_client.ApiClient()).resources.get(
        group=ORBIT_API_GROUP, api_version=ORBIT_API_VERSION, kind=USERSPACE_CR_KIND
    )

    manage_user_namespace(event, api_CoreV1, userspace_dc)


def run_command(cmd: str) -> None:
    """Module to run shell commands."""
    cmds = cmd.split(" ")
    try:
        output = subprocess.run(cmds, stderr=subprocess.STDOUT, shell=False, timeout=120, universal_newlines=True)
        print(output)
    except subprocess.CalledProcessError as exc:
        # print("Command failed with exit code {}, stderr: {}".format(exc.returncode, exc.output.decode("utf-8")))
        raise exc


def create_kubeconfig() -> None:
    logger.info(f"Generating kubeconfig in {KUBECONFIG_PATH}")
    run_command(
        (
            "aws eks update-kubeconfig "
            f"--name orbit-{ORBIT_ENV} "
            f"--role-arn arn:aws:iam::{ACCOUNT_ID}:role{ROLE_PREFIX}orbit-{ORBIT_ENV}-{REGION}-admin "
            f"--kubeconfig {KUBECONFIG_PATH}"
        )
    )

    logger.info("Loading kubeconfig")
    try:
        config.load_kube_config(KUBECONFIG_PATH)
        logger.info("Loaded kubeconfig successfully")
    except config.ConfigException:
        raise Exception("Could not configure kubernetes python client")


def create_userspace(
    userspace_dc: dynamic.DynamicClient,
    name: str,
    env: str,
    space: str,
    team: str,
    user: str,
    team_efsid: str,
    user_email: str,
    owner_reference: Optional[Dict[str, str]] = None,
    labels: Optional[Dict[str, str]] = None,
    annnotations: Optional[Dict[str, str]] = None,
) -> None:
    userspace: Dict[str, Any] = {
        "apiVersion": f"{ORBIT_API_GROUP}/{ORBIT_API_VERSION}",
        "kind": USERSPACE_CR_KIND,
        "metadata": {"name": name, "labels": labels, "annotations": annnotations, "namespace": name},
        "spec": {
            "env": env,
            "space": space,
            "team": team,
            "user": user,
            "teamEfsId": team_efsid,
            "userEmail": user_email,
        },
    }
    if owner_reference is not None:
        userspace["metadata"]["ownerReferences"] = [owner_reference]
    logger.info(f"userspace={userspace}")
    userspace_dc.create(namespace=name, body=userspace)


def create_user_namespace(
    api: client.CoreV1Api,
    userspace_dc: dynamic.DynamicClient,
    user_name: str,
    user_email: str,
    expected_user_namespaces: Dict[str, str],
    namespaces: List[str],
) -> None:
    env = os.environ.get("ORBIT_ENV", "")
    if not env:
        raise ValueError("Orbit Environment ORBIT_ENV is required")
    for team, user_ns in expected_user_namespaces.items():
        try:
            team_namespace = api.read_namespace(name=team).to_dict()
            team_uid = team_namespace.get("metadata", {}).get("uid", None)
            logger.info(f"Retrieved Team Namespace uid: {team_uid}")
        except Exception:
            logger.exception("Error retrieving Team Namespace")
            team_uid = None
        if user_ns not in namespaces:
            logger.info(f"User namespace {user_ns} doesnt exist. Creating...")
            kwargs = {
                "name": user_ns,
                "annotations": {"owner": user_email},
                "labels": {
                    "orbit/efs-id": EFS_FS_ID,
                    "orbit/env": os.environ.get("ORBIT_ENV"),
                    "orbit/space": "user",
                    "orbit/team": team,
                    "orbit/user": user_name,
                    # "istio-injection": "enabled",
                },
            }
            if team_uid:
                kwargs["owner_references"] = [
                    client.V1OwnerReference(api_version="v1", kind="Namespace", name=team, uid=team_uid)
                ]

            body = client.V1Namespace()
            body.metadata = client.V1ObjectMeta(**kwargs)

            try:
                # create userspace namespace resource
                api.create_namespace(body=body)
                logger.info(f"Created namespace {user_ns}")
            except ApiException as ae:
                logger.warning(f"Exception when trying to create user namespace {user_ns}")
                logger.warning(ae.body)

            try:
                # create userspace custom resource for the given user namespace
                logger.info(f"Creating userspace custom resource {user_ns}")
                create_userspace(
                    userspace_dc=userspace_dc,
                    name=user_ns,
                    env=env,
                    space="user",
                    team=team,
                    user=user_name,
                    team_efsid=EFS_FS_ID,
                    user_email=user_email,
                )
                logger.info(f"Created userspace custom resource {user_ns}")
            except ApiException as ae:
                logger.warning(f"Exception when trying to create userspace custom resource {user_ns}")
                logger.warning(ae.body)


def delete_user_namespace(
    api: client.CoreV1Api,
    userspace_dc: dynamic.DynamicClient,
    user_name: str,
    expected_user_namespaces: Dict[str, str],
    namespaces: List[str],
) -> None:
    for user_ns in namespaces:
        if user_ns not in expected_user_namespaces.values():
            delete_user_profile(user_profile=user_ns)
            logger.info(f"User {user_name} is not expected to be part of the {user_ns} namespace. Removing...")
            try:
                userspace_dc.delete(name=user_ns, namespace=user_ns)
                logger.info(f"Removed userspace custom resource {user_ns}")
            except ApiException as ae:
                logger.warning(f"Exception when trying to remove userspace custom resource {user_ns}")
                logger.warning(ae.body)
            try:
                api.delete_collection_namespaced_pod(namespace=user_ns, grace_period_seconds=0)
                api.delete_namespace(name=user_ns)
                logger.info(f"Removed namespace {user_ns}")
            except ApiException as ae:
                logger.warning(f"Exception when trying to remove user namespace {user_ns}")
                logger.warning(ae.body)


def delete_user_profile(user_profile: str) -> None:
    logger.info(f"Removing profile {user_profile}")
    run_command(f"kubectl delete profile {user_profile} --kubeconfig {KUBECONFIG_PATH}")
    time.sleep(5)


def manage_user_namespace(event: Dict[str, Any], api: client.CoreV1Api, userspace_dc: dynamic.DynamicClient) -> None:
    user_name = cast(str, event.get("user_name"))
    user_email = cast(str, event.get("user_email"))
    expected_user_namespaces = cast(Dict[str, str], event.get("expected_user_namespaces"))

    all_ns_raw = api.list_namespace().to_dict()
    all_ns = [
        item.get("metadata").get("name")
        for item in all_ns_raw["items"]
        if item.get("metadata", {}).get("name") and item.get("metadata", {}).get("name").endswith(user_name)
    ]

    create_user_namespace(
        api=api,
        userspace_dc=userspace_dc,
        user_name=user_name,
        user_email=user_email,
        expected_user_namespaces=expected_user_namespaces,
        namespaces=all_ns,
    )

    delete_user_namespace(
        api=api,
        userspace_dc=userspace_dc,
        user_name=user_name,
        expected_user_namespaces=expected_user_namespaces,
        namespaces=all_ns,
    )



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import os
from io import open
from typing import Dict

from setuptools import find_packages, setup

here = os.path.abspath(os.path.dirname(__file__))
about: Dict[str, str] = {}
path = os.path.join(here, "aws_orbit", "__metadata__.py")
with open(file=path, mode="r", encoding="utf-8") as f:
    exec(f.read(), about)

with open("VERSION", "r") as version_file:
    version = version_file.read().strip()

setup(
    name=about["__title__"],
    version=version,
    author="AWS Professional Services",
    author_email="aws-proserve-opensource@amazon.com",
    url="https://github.com/awslabs/aws-orbit-workbench",
    project_urls={"Org Site": "https://aws.amazon.com/professional-services/"},
    description=about["__description__"],
    license=about["__license__"],
    packages=find_packages(include=["aws_orbit", "aws_orbit.*"]),
    keywords=["aws", "cdk"],
    python_requires=">=3.7, <3.9",
    install_requires=[
        "boto3~=1.18.0",
        "botocore~=1.15",
        "PyYAML~=5.4",
        "click~=7.1.0",
        "tqdm~=4.56.0",
        "pyyaml-include~=1.2",
        "jsonpath-ng~=1.5.0",
        "marshmallow_dataclass[enum,union]~=8.3.0",
        "cfn-flip~=1.2.3",
        "python-slugify~=4.0.1",
        "kubernetes~=12.0.1",
    ],
    entry_points={"console_scripts": ["orbit = aws_orbit.__main__:main"]},
    classifiers=[
        "Development Status :: 4 - Beta",
        "Framework :: Jupyter",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Typing :: Typed",
    ],
    include_package_data=True,
    extras_require={"utils": [f"aws-orbit-sdk~={version}"]},
)



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

# flake8: noqa: F811

import logging
from typing import ClassVar, Dict, List, Optional, Type

import botocore.exceptions
from dataclasses import field
from marshmallow import Schema
from marshmallow_dataclass import dataclass

from aws_orbit.models.common import BaseSchema
from aws_orbit.utils import boto3_client

_logger: logging.Logger = logging.getLogger(__name__)


@dataclass(base_schema=BaseSchema, frozen=True)
class UserIdGroupPair:
    description: str
    group_id: str


@dataclass(base_schema=BaseSchema, frozen=True)
class IpPermission:
    Schema: ClassVar[Type[Schema]] = Schema
    from_port: int
    to_port: int
    ip_protocol: str
    user_id_group_pairs: List[UserIdGroupPair] = field(default_factory=list)


def authorize_security_group_ingress(group_id: str, ip_permissions: List[IpPermission]) -> None:
    permissions = [IpPermission.Schema().dump(i) for i in ip_permissions]
    _logger.debug(f"Authorizing Ingress for Security Group: {group_id}\nPermissions: {permissions}")

    ec2_client = boto3_client("ec2")
    try:
        ec2_client.authorize_security_group_ingress(GroupId=group_id, IpPermissions=permissions)
    except botocore.exceptions.ClientError as ex:
        _logger.error("Error Authorizing Ingress", ex)
        if ex.response.get("Error", {}).get("Code", "Unknown") != "InvalidPermission.Duplicate":
            raise
        else:
            _logger.debug("Ingress previously authorized")


def get_az_from_subnet(subnets: List[str]) -> Dict[str, str]:
    ec2_client = boto3_client("ec2")
    _logger.info("Subnets info: %s", subnets)
    az_subnet_map = {}
    try:
        response = ec2_client.describe_subnets(SubnetIds=subnets)
        az_subnet_map = {entry["SubnetId"]: entry["AvailabilityZone"] for entry in response["Subnets"]}
    except botocore.exceptions.ClientError as ex:
        _logger.error("Error Describing Subnets: %s", ex)
        if ex.response.get("Error", {}).get("Code", "Unknown") != "InvalidSubnetID.NotFound":
            raise
        else:
            _logger.debug("Exception caught while describing subnets: %s", ex)
    return az_subnet_map


def authorize_security_group_egress(group_id: str, ip_permissions: List[IpPermission]) -> None:
    permissions = [IpPermission.Schema().dump(i) for i in ip_permissions]
    _logger.debug(f"Authorizing Egress for Security Group: {group_id}\nPermissions: {permissions}")

    ec2_client = boto3_client("ec2")
    try:
        ec2_client.authorize_security_group_egress(GroupId=group_id, IpPermissions=permissions)
    except botocore.exceptions.ClientError as ex:
        _logger.error("Error Authorizing Egress", ex)
        if ex.response.get("Error", {}).get("Code", "Unknown") != "InvalidPermission.Duplicate":
            raise
        else:
            _logger.debug("Egress previously authorized")


def revoke_security_group_ingress(group_id: str, ip_permissions: List[IpPermission]) -> None:
    permissions = [IpPermission.Schema().dump(i) for i in ip_permissions]
    _logger.debug(f"Revoking Ingress for Security Group: {group_id}\nPermissions: {permissions}")

    ec2_client = boto3_client("ec2")
    try:
        ec2_client.revoke_security_group_ingress(GroupId=group_id, IpPermissions=permissions)
    except botocore.exceptions.ClientError as ex:
        _logger.error("Error Revoking Ingress", ex)
        if ex.response.get("Error", {}).get("Code", "Unknown") != "InvalidPermission.NotFound":
            raise
        else:
            _logger.debug("Ingress not previously authorized")
    except botocore.exceptions.ParamValidationError as err:
        _logger.error(f"Error revoking ingress, parameter validations failed: {err}")


def revoke_security_group_egress(group_id: str, ip_permissions: List[IpPermission]) -> None:
    permissions = [IpPermission.Schema().dump(i) for i in ip_permissions]
    _logger.debug(f"Revoking Egress for Security Group: {group_id}\nPermissions: {permissions}")

    ec2_client = boto3_client("ec2")
    try:
        ec2_client.revoke_security_group_egress(GroupId=group_id, IpPermissions=permissions)
    except botocore.exceptions.ClientError as ex:
        _logger.error("Error Revoking Egress", ex)
        if ex.response.get("Error", {}).get("Code", "Unknown") != "InvalidPermission.NotFound":
            raise
        else:
            _logger.debug("Egress not previously authorized")
    except botocore.exceptions.ParamValidationError as err:
        _logger.error(f"Error revoking engress, parameter validations failed: {err}")



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import logging
import os
from typing import TYPE_CHECKING, List, Optional, cast

import boto3

from aws_orbit import ORBIT_CLI_ROOT, cdk, cleanup, docker
from aws_orbit.services import cfn, ecr, iam, ssm

if TYPE_CHECKING:
    from aws_orbit.models.changeset import ListChangeset
    from aws_orbit.models.context import Context

_logger: logging.Logger = logging.getLogger(__name__)


def update_subnet_tags(context: "Context") -> None:
    ec2 = boto3.client("ec2")
    cluster_name = f"orbit-{context.name}"
    if context.networking.public_subnets:
        subnet_ids = [x.subnet_id for x in context.networking.public_subnets]
        _logger.debug('updating public subnet tags with "kubernetes.io/role/elb"')
        ec2.create_tags(
            Resources=subnet_ids,
            Tags=[
                {"Key": "kubernetes.io/role/elb", "Value": "1"},
                {"Key": f"kubernetes.io/cluster/{cluster_name}", "Value": "shared"},
            ],
        )

    subnet_ids = [x for x in context.networking.data.nodes_subnets]
    _logger.debug('updating private subnet tags with "kubernetes.io/role/internal-elb"')
    ec2.create_tags(
        Resources=subnet_ids,
        Tags=[
            {"Key": f"kubernetes.io/cluster/{cluster_name}", "Value": "shared"},
            {"Key": "kubernetes.io/role/internal-elb", "Value": "1"},
        ],
    )


def deploy(
    context: "Context",
    eks_system_masters_roles_changes: Optional["ListChangeset"],
) -> None:
    _logger.debug("Stack name: %s", context.env_stack_name)

    if eks_system_masters_roles_changes and (
        eks_system_masters_roles_changes.added_values or eks_system_masters_roles_changes.removed_values
    ):
        iam.update_assume_role_roles(
            account_id=context.account_id,
            role_name=cast(str, context.toolkit.admin_role),
            roles_to_add=eks_system_masters_roles_changes.added_values,
            roles_to_remove=eks_system_masters_roles_changes.removed_values,
        )

    args: List[str] = [context.name]

    update_subnet_tags(context=context)

    cdk.deploy(
        context=context,
        stack_name=context.env_stack_name,
        app_filename=os.path.join(ORBIT_CLI_ROOT, "remote_files", "cdk", "env.py"),
        args=args,
    )
    context.fetch_env_data()


def destroy(context: "Context") -> None:
    _logger.debug("Stack name: %s", context.env_stack_name)
    if cfn.does_stack_exist(stack_name=context.env_stack_name):
        docker.login(context=context)
        _logger.debug("DockerHub and ECR Logged in")
        ecr.cleanup_remaining_repos(env_name=context.name)
        args = [context.name]
        cdk.destroy(
            context=context,
            stack_name=context.env_stack_name,
            app_filename=os.path.join(ORBIT_CLI_ROOT, "remote_files", "cdk", "env.py"),
            args=args,
        )
        cleanup.delete_kubeflow_roles(context.env_stack_name, context.region, context.account_id)
        ssm.cleanup_context(env_name=context.name)



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import concurrent.futures
import json
import logging
import os
import subprocess
from concurrent.futures import Future
from typing import Any, List, Optional, Tuple, cast

from aws_orbit import bundle, docker, plugins, remote, sh
from aws_orbit.models.changeset import Changeset, load_changeset_from_ssm
from aws_orbit.models.context import Context, ContextSerDe, FoundationContext, TeamContext
from aws_orbit.models.manifest import ImageManifest, ImagesManifest, Manifest, ManifestSerDe
from aws_orbit.remote_files import cdk_toolkit, eksctl, env, foundation, helm, kubectl, teams, utils
from aws_orbit.services import codebuild, ecr, kms, secretsmanager
from aws_orbit.utils import boto3_client

_logger: logging.Logger = logging.getLogger(__name__)


def _deploy_image(args: Tuple[str, ...]) -> None:
    _logger.debug("_deploy_image args: %s", args)
    if len(args) < 4:
        raise ValueError("Unexpected number of values in args.")
    env: str = args[0]
    image_name: str = args[1]
    dir: str = args[2]
    script: Optional[str] = args[3] if args[3] != "NO_SCRIPT" else None
    build_args = args[4:]

    manifest = ManifestSerDe.load_manifest_from_ssm(env_name=env, type=Manifest)
    context: "Context" = ContextSerDe.load_context_from_ssm(env_name=env, type=Context)
    _logger.debug("context: %s", vars(context))

    if manifest is None:
        raise Exception("Unable to load required Manifest from SSM")

    docker.login(context=context)
    _logger.debug("DockerHub and ECR Logged in")
    _logger.debug("Deploying the %s Docker image", image_name)

    ecr_repo = f"orbit-{context.name}/{image_name}"
    if not ecr.describe_repositories(repository_names=[ecr_repo]):
        ecr.create_repository(repository_name=ecr_repo, env_name=context.name)

    image_def: ImageManifest = getattr(manifest.images, image_name.replace("-", "_"))
    source = image_def.get_source(account_id=context.account_id, region=context.region)
    if source == "code":
        _logger.debug("Building and deploy docker image from source...")
        path = os.path.join(os.getcwd(), "bundle", dir)
        _logger.debug("path: %s", path)
        if script is not None:
            sh.run(f"sh {script}", cwd=path)
        tag = cast(str, image_def.version)
        docker.deploy_image_from_source(
            context=context,
            dir=path,
            name=ecr_repo,
            build_args=cast(Optional[List[str]], build_args),
            tag=tag,
        )
    else:
        _logger.debug("Replicating docker image to ECR...")
        docker.replicate_image(
            context=context,
            image_name=image_name,
            deployed_name=ecr_repo,
            source=source,
            source_repository=image_def.repository,
            source_version=image_def.version,
        )

    _logger.debug("Docker Image Deployed to ECR")


def _deploy_image_remotely(context: "Context", name: str, bundle_path: str, buildspec: codebuild.SPEC_TYPE) -> None:
    _logger.debug("Deploying %s Docker Image remotely into ECR", name)
    remote.run(
        command_name=f"deploy_image-{name}",
        context=context,
        bundle_path=bundle_path,
        buildspec=buildspec,
        codebuild_log_callback=None,
        timeout=30,
    )
    _logger.debug("%s Docker Image deployed into ECR", name)


def _deploy_images_batch(
    manifest: Manifest, context: "Context", images: List[Tuple[str, Optional[str], Optional[str], List[str]]]
) -> None:
    _logger.debug("images:\n%s", images)

    new_images_manifest = {name: getattr(context.images, name) for name in context.images.names}
    max_workers = 5
    with concurrent.futures.ThreadPoolExecutor(max_workers=max_workers) as executor:
        futures: List[Future[Any]] = []
        name: str = ""
        dir: Optional[str] = None
        script: Optional[str] = None
        build_args: List[str] = []

        for name, dir, script, build_args in images:
            _logger.debug("name: %s | script: %s", name, script)

            path = os.path.join(os.getcwd(), "bundle", name)
            _logger.debug("path: %s", path)

            image_attr_name = name.replace("-", "_")
            image_def: ImageManifest = getattr(manifest.images, image_attr_name)
            tag = image_def.version
            if image_def.get_source(account_id=context.account_id, region=context.region) == "code":
                dirs: List[Tuple[str, str]] = [(path, cast(str, dir))]
            else:
                dirs = []

            bundle_path = bundle.generate_bundle(command_name=f"deploy_image-{name}", context=context, dirs=dirs)
            _logger.debug("bundle_path: %s", bundle_path)
            script_str = "NO_SCRIPT" if script is None else script
            build_args = [] if build_args is None else build_args
            buildspec = codebuild.generate_spec(
                context=context,
                plugins=False,
                cmds_build=[
                    "orbit remote --command _deploy_image "
                    f"{context.name} {name} {dir} {script_str} {' '.join(build_args)}"
                ],
            )
            new_images_manifest[image_attr_name] = ImageManifest(
                repository=f"{context.account_id}.dkr.ecr.{context.region}.amazonaws.com/orbit-{context.name}/{name}",
                version=tag,
                path=None,
            )
            futures.append(executor.submit(_deploy_image_remotely, context, name, bundle_path, buildspec))

        for f in futures:
            f.result()

        context.images = ImagesManifest(**new_images_manifest)
        ContextSerDe.dump_context_to_ssm(context=context)


def deploy_images_remotely(manifest: Manifest, context: "Context", skip_images: bool = True) -> None:
    # Required images we always build/replicate
    images: List[Tuple[str, Optional[str], Optional[str], List[str]]] = []

    # This isn't obvious to read, but when skipping image builds, we intentionally remove these images if the source
    # is code. Otherwise, we build/deploy them
    if not (manifest.images.k8s_utilities.get_source(context.account_id, context.region) == "code" and skip_images):
        images.append(("k8s-utilities", "k8s-utilities", None, []))
    if not (manifest.images.orbit_controller.get_source(context.account_id, context.region) == "code" and skip_images):
        images.append(("orbit-controller", "orbit-controller", None, []))

    # Secondary images we can optionally skip
    if not skip_images:
        # When not skipping images, include these secondaries if their source isn't the default ecr-public
        if manifest.images.jupyter_user.get_source(context.account_id, context.region) != "ecr-public":
            images.append(("jupyter-user", "jupyter-user", "build.sh", []))

    _logger.debug("Building/repclicating Container Images")
    _deploy_images_batch(manifest=manifest, context=context, images=images)


def deploy_credentials(args: Tuple[str, ...]) -> None:
    _logger.debug("args: %s", args)
    if len(args) != 2:
        raise ValueError("Unexpected number of values in args")
    env_name: str = args[0]
    ciphertext: str = args[1]
    context: "Context" = ContextSerDe.load_context_from_ssm(env_name=env_name, type=Context)
    _logger.debug("Context loaded.")

    new_credentials = json.loads(kms.decrypt(context=context, ciphertext=ciphertext))
    secret_id = f"orbit-{env_name}-docker-credentials"
    existing_credentials = secretsmanager.get_secret_value(secret_id=secret_id)
    for registry, creds in new_credentials.items():
        username = creds.get("username", "")
        password = creds.get("password", "")
        try:
            subprocess.check_call(
                f"docker login --username '{username}' --password '{password}' {registry}", shell=True
            )
        except Exception as e:
            _logger.error("Invalid Registry Credentials")
            _logger.exception(e)
            return
        else:
            existing_credentials = {**existing_credentials, **new_credentials}
    secretsmanager.put_secret_value(secret_id=secret_id, secret=existing_credentials)
    _logger.debug("Registry Credentials deployed")


def deploy_foundation(args: Tuple[str, ...]) -> None:
    _logger.debug("args: %s", args)
    if len(args) != 1:
        raise ValueError("Unexpected number of values in args")
    env_name: str = args[0]
    context: "FoundationContext" = ContextSerDe.load_context_from_ssm(env_name=env_name, type=FoundationContext)
    _logger.debug("Context loaded.")
    docker.login(context=context)
    _logger.debug("DockerHub and ECR Logged in")
    cdk_toolkit.deploy(context=context)
    _logger.debug("CDK Toolkit Stack deployed")
    foundation.deploy(context=context)
    _logger.debug("Demo Stack deployed")


def deploy_env(args: Tuple[str, ...]) -> None:
    _logger.debug("args: %s", args)
    if len(args) == 2:
        env_name: str = args[0]
        skip_images_remote_flag: str = str(args[1])
    else:
        raise ValueError("Unexpected number of values in args")

    manifest: Optional[Manifest] = ManifestSerDe.load_manifest_from_ssm(env_name=env_name, type=Manifest)
    _logger.debug("Manifest loaded.")
    context: "Context" = ContextSerDe.load_context_from_ssm(env_name=env_name, type=Context)
    _logger.debug("Context loaded.")
    changeset: Optional["Changeset"] = load_changeset_from_ssm(env_name=env_name)
    _logger.debug("Changeset loaded.")

    if manifest is None:
        raise Exception("Unable to load Manifest")

    docker.login(context=context)
    _logger.debug("DockerHub and ECR Logged in")
    cdk_toolkit.deploy(context=context)
    _logger.debug("CDK Toolkit Stack deployed")
    env.deploy(
        context=context,
        eks_system_masters_roles_changes=changeset.eks_system_masters_roles_changeset if changeset else None,
    )

    _logger.debug("Env Stack deployed")
    deploy_images_remotely(manifest=manifest, context=context, skip_images=skip_images_remote_flag == "skip-images")
    _logger.debug("Docker Images deployed")
    eksctl.deploy_env(
        context=context,
        changeset=changeset,
    )
    _logger.debug("EKS Environment Stack deployed")
    kubectl.deploy_env(context=context)
    _logger.debug("Kubernetes Environment components deployed")

    helm.deploy_env(context=context)
    _logger.debug("Helm Charts installed")

    k8s_context = utils.get_k8s_context(context=context)
    kubectl.fetch_kubectl_data(context=context, k8s_context=k8s_context)
    ContextSerDe.dump_context_to_ssm(context=context)
    _logger.debug("Updating userpool redirect")
    _update_userpool_client(context=context)
    _update_userpool(context=context)


def _update_userpool(context: Context) -> None:
    cognito_client = boto3_client("cognito-idp")

    function_arn = (
        f"arn:aws:lambda:{context.region}:{context.account_id}:function:orbit-{context.name}-post-authentication"
    )

    cognito_client.update_user_pool(
        UserPoolId=context.user_pool_id,
        LambdaConfig={"PostAuthentication": function_arn, "PostConfirmation": function_arn},
    )


def _update_userpool_client(context: Context) -> None:
    cognito = boto3_client("cognito-idp")
    cognito.update_user_pool_client(
        UserPoolId=context.user_pool_id,
        ClientId=context.user_pool_client_id,
        CallbackURLs=[
            f"{context.landing_page_url}/oauth2/idpresponse",
            f"{context.landing_page_url}/orbit/login",
            f"{context.landing_page_url}/saml2/idpresponse",
        ],
        LogoutURLs=[
            f"{context.landing_page_url}/orbit/logout",
        ],
        ExplicitAuthFlows=["ALLOW_CUSTOM_AUTH", "ALLOW_USER_SRP_AUTH", "ALLOW_REFRESH_TOKEN_AUTH"],
        SupportedIdentityProviders=["COGNITO"],
        AllowedOAuthFlows=["code"],
        AllowedOAuthScopes=["aws.cognito.signin.user.admin", "email", "openid", "profile"],
        AllowedOAuthFlowsUserPoolClient=True,
        PreventUserExistenceErrors="ENABLED",
        TokenValidityUnits={"AccessToken": "minutes", "IdToken": "minutes", "RefreshToken": "days"},
        RefreshTokenValidity=90,
        AccessTokenValidity=60,
        IdTokenValidity=60,
    )


def deploy_teams(args: Tuple[str, ...]) -> None:
    _logger.debug("args: %s", args)
    if len(args) == 1:
        env_name: str = args[0]
    else:
        raise ValueError("Unexpected number of values in args")

    context: "Context" = ContextSerDe.load_context_from_ssm(env_name=env_name, type=Context)
    _logger.debug("Context loaded.")
    changeset: Optional["Changeset"] = load_changeset_from_ssm(env_name=env_name)
    _logger.debug("Changeset loaded.")

    if changeset:
        plugins.PLUGINS_REGISTRIES.load_plugins(
            context=context, plugin_changesets=changeset.plugin_changesets, teams_changeset=changeset.teams_changeset
        )
        _logger.debug("Plugins loaded")

    docker.login(context=context)
    _logger.debug("DockerHub and ECR Logged in")
    if changeset and changeset.teams_changeset and changeset.teams_changeset.removed_teams_names:
        kubectl.write_kubeconfig(context=context)
        for team_name in changeset.teams_changeset.removed_teams_names:
            team_context: Optional["TeamContext"] = context.get_team_by_name(name=team_name)
            if team_context is None:
                raise RuntimeError(f"TeamContext {team_name} not found!")
            _logger.debug("Destory all user namespaces for %s", team_context.name)
            sh.run(f"kubectl delete namespaces -l orbit/team={team_context.name},orbit/space=user --wait=true")
            _logger.debug("Destroying team %s", team_name)
            plugins.PLUGINS_REGISTRIES.destroy_team_plugins(context=context, team_context=team_context)
            _logger.debug("Team Plugins destroyed")
            helm.destroy_team(context=context, team_context=team_context)
            _logger.debug("Team Helm Charts uninstalled")
            kubectl.destroy_team(context=context, team_context=team_context)
            _logger.debug("Kubernetes Team components destroyed")
            eksctl.destroy_team(context=context, team_context=team_context)
            _logger.debug("EKS Team Stack destroyed")
            teams.destroy_team(context=context, team_context=team_context)
            _logger.debug("Team %s destroyed", team_name)
            context.remove_team_by_name(name=team_name)
            ContextSerDe.dump_context_to_ssm(context=context)

    team_names = [t.name for t in context.teams]
    if changeset and changeset.teams_changeset and changeset.teams_changeset.added_teams_names:
        team_names.extend(changeset.teams_changeset.added_teams_names)

    manifest: Optional["Manifest"] = ManifestSerDe.load_manifest_from_ssm(env_name=context.name, type=Manifest)
    if manifest is None:
        raise RuntimeError(f"Manifest {context.name} not found!")
    kubectl.write_kubeconfig(context=context)
    for team_name in team_names:
        team_manifest = manifest.get_team_by_name(name=team_name)
        if team_manifest is None:
            raise RuntimeError(f"TeamManifest {team_name} not found!")
        teams.deploy_team(context=context, manifest=manifest, team_manifest=team_manifest)
        _logger.debug("Team Stacks deployed")
        team_context = context.get_team_by_name(name=team_name)
        if team_context is None:
            raise RuntimeError(f"TeamContext {team_name} not found!")
        eksctl.deploy_team(context=context, team_context=team_context)
        _logger.debug("EKS Team Stack deployed")
        kubectl.deploy_team(context=context, team_context=team_context)
        _logger.debug("Kubernetes Team components deployed")
        helm.deploy_team(context=context, team_context=team_context)
        _logger.debug("Team Helm Charts installed")
        plugins.PLUGINS_REGISTRIES.deploy_team_plugins(
            context=context, team_context=team_context, changes=changeset.plugin_changesets if changeset else []
        )

        team_context.plugins = team_manifest.plugins
        ContextSerDe.dump_context_to_ssm(context=context)
        _logger.debug("Team Plugins deployed")

    _logger.debug("Teams deployed")



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

# flake8: noqa: F811

import logging
import os
from copy import deepcopy
from typing import TYPE_CHECKING, Any, ClassVar, Dict, List, Optional, Set, Type, cast

import yaml
from dataclasses import field
from marshmallow import Schema
from marshmallow_dataclass import dataclass

from aws_orbit.models.common import BaseSchema
from aws_orbit.models.manifest import ManagedNodeGroupManifest
from aws_orbit.services import ssm

if TYPE_CHECKING:
    from aws_orbit.messages import MessagesContext
    from aws_orbit.models.context import Context, TeamContext
    from aws_orbit.models.manifest import Manifest, TeamManifest

_logger: logging.Logger = logging.getLogger(__name__)


@dataclass(base_schema=BaseSchema)
class TeamsChangeset:
    Schema: ClassVar[Type[Schema]] = Schema
    removed_teams_names: List[str]
    added_teams_names: List[str]


@dataclass(base_schema=BaseSchema)
class PluginChangeset:
    Schema: ClassVar[Type[Schema]] = Schema
    team_name: str
    old: List[str] = field(default_factory=list)
    old_paths: Dict[str, Optional[str]] = field(default_factory=dict)
    old_parameters: Dict[str, Dict[str, Any]] = field(default_factory=dict)
    old_modules: Dict[str, str] = field(default_factory=dict)
    new: List[str] = field(default_factory=list)
    new_paths: Dict[str, Optional[str]] = field(default_factory=dict)
    new_parameters: Dict[str, Dict[str, Any]] = field(default_factory=dict)
    new_modules: Dict[str, str] = field(default_factory=dict)


@dataclass(base_schema=BaseSchema)
class ImageChangeset:
    Schema: ClassVar[Type[Schema]] = Schema
    team_name: str
    old_image: Optional[str]
    new_image: Optional[str]


@dataclass(base_schema=BaseSchema)
class ExternalIDPChangeset:
    Schema: ClassVar[Type[Schema]] = Schema
    old_provider: Optional[str]
    new_provider: Optional[str]
    old_label: Optional[str]
    new_label: Optional[str]


@dataclass(base_schema=BaseSchema)
class ListChangeset:
    Schema: ClassVar[Type[Schema]] = Schema
    removed_values: Optional[List[Any]] = None
    added_values: Optional[List[Any]] = None


@dataclass(base_schema=BaseSchema)
class ManagedNodeGroupsChangeset:
    Schema: ClassVar[Type[Schema]] = Schema
    removed_nodegroups: List[ManagedNodeGroupManifest]
    added_nodegroups: List[ManagedNodeGroupManifest]
    modified_nodegroups: List[ManagedNodeGroupManifest]


@dataclass(base_schema=BaseSchema)
class Changeset:
    Schema: ClassVar[Type[Schema]] = Schema
    image_changesets: List[ImageChangeset]
    plugin_changesets: List[PluginChangeset]
    external_idp_changeset: Optional[ExternalIDPChangeset]
    teams_changeset: Optional[TeamsChangeset]
    eks_system_masters_roles_changeset: Optional[ListChangeset]
    managed_nodegroups_changeset: Optional[ManagedNodeGroupsChangeset]


def _check_images(manifest: "Manifest", context: "Context", msg_ctx: "MessagesContext") -> List[ImageChangeset]:
    image_changesets: List[ImageChangeset] = []
    for team in context.teams:
        new_team: Optional["TeamManifest"] = manifest.get_team_by_name(team.name)
        if new_team is None:
            continue
        _logger.debug("Inpecting Image Change for team %s: %s -> %s", team.name, team.image, new_team.image)
        if team.image != new_team.image:
            msg_ctx.info(f"Image change detected for Team {team.name}: {team.image} -> {new_team.image}")
            image_changesets.append(ImageChangeset(team_name=team.name, old_image=team.image, new_image=new_team.image))
    return image_changesets


def _check_external_idp(
    manifest: "Manifest", context: "Context", msg_ctx: "MessagesContext"
) -> Optional[ExternalIDPChangeset]:
    _logger.debug("Inpecting External IDP Change...")
    old_provider: Optional[str] = context.cognito_external_provider
    new_provider: Optional[str] = manifest.cognito_external_provider
    _logger.debug("Provider: %s -> %s", old_provider, new_provider)
    old_label: Optional[str] = context.cognito_external_provider_label
    new_label: Optional[str] = manifest.cognito_external_provider_label
    _logger.debug("Label: %s -> %s", old_label, new_label)
    if old_provider != new_provider or old_label != new_label:
        external_idp_changeset: Optional[ExternalIDPChangeset] = ExternalIDPChangeset(
            old_provider=old_provider, new_provider=new_provider, old_label=old_label, new_label=new_label
        )
        msg_ctx.info(f"External IDP change detected: {old_provider} ({old_label}) -> {new_provider} ({new_label})")
    else:
        external_idp_changeset = None
    return external_idp_changeset


def _check_teams(manifest: "Manifest", context: "Context", msg_ctx: "MessagesContext") -> Optional[TeamsChangeset]:
    _logger.debug("Inpecting Teams changes...")
    old_teams: List["TeamContext"] = deepcopy(context.teams)
    old_names = sorted([t.name for t in old_teams])
    new_teams: List["TeamManifest"] = deepcopy(manifest.teams)
    new_names: List[str] = sorted([t.name for t in new_teams])
    _logger.debug("Teams: %s -> %s", old_names, new_names)
    removed_teams: Set[str] = set(old_names) - set(new_names)
    _logger.debug("removed_teams: %s", removed_teams)
    added_teams: Set[str] = set(new_names) - set(old_names)
    _logger.debug("added_teams: %s", added_teams)
    if removed_teams or added_teams:
        teams_changeset: Optional[TeamsChangeset] = TeamsChangeset(
            removed_teams_names=list(removed_teams),
            added_teams_names=list(added_teams),
        )
        msg_ctx.info(f"Removed teams: {list(removed_teams)}")
        msg_ctx.info(f"Added teams: {list(added_teams)}")
    else:
        teams_changeset = None
    return teams_changeset


def _check_team_plugins(
    team_manifest: "TeamManifest", context: "Context", msg_ctx: "MessagesContext"
) -> Optional[PluginChangeset]:
    new_names: List[str] = [p.plugin_id for p in team_manifest.plugins]
    old_team: Optional["TeamContext"] = context.get_team_by_name(name=team_manifest.name)
    if old_team:
        old_names: List[str] = [p.plugin_id for p in old_team.plugins]
    else:
        old_names = []
    _logger.debug("Inpecting Plugins Change for team %s: %s -> %s", team_manifest.name, old_names, new_names)

    if sorted(old_names) != sorted(new_names):
        msg_ctx.info(f"Plugin change detected for Team {team_manifest.name}: {old_names} -> {new_names}")
        return PluginChangeset(
            team_name=team_manifest.name,
            old=old_names,
            old_paths={p.plugin_id: p.path for p in old_team.plugins} if old_team else {},
            old_parameters={p.plugin_id: p.parameters for p in old_team.plugins} if old_team else {},
            old_modules={p.plugin_id: p.module for p in old_team.plugins} if old_team else {},
            new=new_names,
            new_paths={p.plugin_id: p.path for p in team_manifest.plugins},
            new_parameters={p.plugin_id: p.parameters for p in team_manifest.plugins},
            new_modules={p.plugin_id: p.module for p in team_manifest.plugins},
        )
    return None


def _get_team_by_name(teams: List["TeamContext"], name: str) -> Optional["TeamContext"]:
    for t in teams:
        if t.name == name:
            return t
    return None


def _check_plugins(
    manifest: "Manifest", context: "Context", msg_ctx: "MessagesContext", teams_changeset: Optional[TeamsChangeset]
) -> List[PluginChangeset]:
    plugin_changesets: List[PluginChangeset] = []

    removed_list: List[str] = teams_changeset.removed_teams_names if teams_changeset else []
    _logger.debug("removed_list: %s", removed_list)
    if teams_changeset and removed_list:  # Removed teams
        removed_teams: List[Optional["TeamContext"]] = [context.get_team_by_name(name=x) for x in removed_list]
        for team in [x for x in removed_teams if x is not None]:
            old_names: List[str] = [p.plugin_id for p in team.plugins]
            if not old_names:
                continue
            plugin_changesets.append(
                PluginChangeset(
                    team_name=team.name,
                    old=old_names,
                    old_paths={p.plugin_id: p.path for p in team.plugins},
                    old_parameters={p.plugin_id: p.parameters for p in team.plugins},
                    old_modules={p.plugin_id: p.module for p in team.plugins},
                )
            )

    for team_manifest in deepcopy(manifest.teams):  # Existing teams into the manifest
        plugin_changeset: Optional[PluginChangeset] = _check_team_plugins(
            team_manifest=team_manifest, context=context, msg_ctx=msg_ctx
        )
        if plugin_changeset is not None:
            plugin_changesets.append(plugin_changeset)

    return plugin_changesets


def _check_eks_system_masters_roles(
    manifest: "Manifest", context: "Context", msg_ctx: "MessagesContext"
) -> Optional[ListChangeset]:
    _logger.debug("Inpecting EKS system:masters Roles changes...")
    old_roles: List[str] = sorted(cast(List[str], context.eks_system_masters_roles))
    new_roles: List[str] = sorted(cast(List[str], manifest.eks_system_masters_roles))
    _logger.debug("Roles: %s -> %s", old_roles, new_roles)
    removed_roles: List[str] = list(set(old_roles) - set(new_roles))
    added_roles: List[str] = list(set(new_roles) - set(old_roles))
    _logger.debug("removed_roles: %s", removed_roles)
    _logger.debug("added_roles: %s", added_roles)
    if removed_roles or added_roles:
        list_changeset: Optional[ListChangeset] = ListChangeset(removed_values=removed_roles, added_values=added_roles)
        msg_ctx.info(f"Removed system:masters Roles: {list(removed_roles)}")
        msg_ctx.info(f"Added system:masters Roles: {list(added_roles)}")
    else:
        list_changeset = None
    return list_changeset


def _check_managed_nodegroups(
    manifest: "Manifest", context: "Context", msg_ctx: "MessagesContext"
) -> Optional[ManagedNodeGroupsChangeset]:
    _logger.debug("Inpecting Managed NodeGroups changes...")
    old_nodegroups: List[str] = sorted([ng.name for ng in context.managed_nodegroups])
    new_nodegroups: List[str] = sorted([ng.name for ng in manifest.managed_nodegroups])
    _logger.debug("ManagedNodeGroups: %s -> %s", old_nodegroups, new_nodegroups)
    removed_nodegroups: List[str] = list(set(old_nodegroups) - set(new_nodegroups))
    added_nodegroups: List[str] = list(set(new_nodegroups) - set(old_nodegroups))

    modified_nodegroups: List[ManagedNodeGroupManifest] = []
    for ng in manifest.managed_nodegroups:
        current_ng = next((ng for ng in context.managed_nodegroups if ng.name == ng.name), None)
        if current_ng:
            if (
                ng.nodes_num_desired != current_ng.nodes_num_desired
                or ng.nodes_num_max != current_ng.nodes_num_max
                or ng.nodes_num_min != current_ng.nodes_num_min
            ):
                modified_nodegroups.append(ng)

    _logger.debug("removed_nodegroups: %s", removed_nodegroups)
    _logger.debug("added_nodegroups: %s", added_nodegroups)
    _logger.debug("modified_nodegroups: %s", [ng.name for ng in modified_nodegroups])

    if removed_nodegroups or added_nodegroups or modified_nodegroups:
        managed_nodegroups_changeset: Optional[ManagedNodeGroupsChangeset] = ManagedNodeGroupsChangeset(
            removed_nodegroups=[ng for ng in context.managed_nodegroups if ng.name in removed_nodegroups],
            added_nodegroups=[ng for ng in manifest.managed_nodegroups if ng.name in added_nodegroups],
            modified_nodegroups=modified_nodegroups,
        )
        msg_ctx.info(f"Removed ManagedNodeGroups: {list(removed_nodegroups)}")
        msg_ctx.info(f"Added ManagedNodeGroups: {list(added_nodegroups)}")
        msg_ctx.info(f"Modified ManagedNodeGroups: {[ng.name for ng in modified_nodegroups]}")
    else:
        managed_nodegroups_changeset = None
    return managed_nodegroups_changeset


def extract_changeset(manifest: "Manifest", context: "Context", msg_ctx: "MessagesContext") -> Changeset:
    image_changesets: List[ImageChangeset] = _check_images(manifest=manifest, context=context, msg_ctx=msg_ctx)
    external_idp_changeset: Optional[ExternalIDPChangeset] = _check_external_idp(
        manifest=manifest, context=context, msg_ctx=msg_ctx
    )
    teams_changeset: Optional[TeamsChangeset] = _check_teams(manifest=manifest, context=context, msg_ctx=msg_ctx)
    plugin_changesets: List[PluginChangeset] = _check_plugins(
        manifest=manifest, context=context, msg_ctx=msg_ctx, teams_changeset=teams_changeset
    )
    eks_system_masters_roles_changeset: Optional[ListChangeset] = _check_eks_system_masters_roles(
        manifest=manifest, context=context, msg_ctx=msg_ctx
    )
    managed_nodegroups_changeset: Optional[ManagedNodeGroupsChangeset] = _check_managed_nodegroups(
        manifest=manifest, context=context, msg_ctx=msg_ctx
    )
    changeset: Changeset = Changeset(
        image_changesets=image_changesets,
        plugin_changesets=plugin_changesets,
        external_idp_changeset=external_idp_changeset,
        teams_changeset=teams_changeset,
        eks_system_masters_roles_changeset=eks_system_masters_roles_changeset,
        managed_nodegroups_changeset=managed_nodegroups_changeset,
    )
    dump_changeset_to_ssm(env_name=context.name, changeset=changeset)
    return changeset


def dump_changeset_to_ssm(env_name: str, changeset: Changeset) -> None:
    _logger.debug("Writing changeset to SSM parameter.")
    manifest_parameter_name: str = f"/orbit/{env_name}/changeset"
    ssm.put_parameter(name=manifest_parameter_name, obj=Changeset.Schema().dump(changeset))


def dump_changeset_to_str(changeset: Changeset) -> str:
    content: Dict[str, Any] = cast(Dict[str, Any], Changeset.Schema().dump(changeset))
    return cast(str, yaml.dump(content, sort_keys=False))


def load_changeset_from_ssm(env_name: str) -> Optional[Changeset]:
    content = ssm.get_parameter_if_exists(name=f"/orbit/{env_name}/changeset")
    if content is None:
        return None
    return cast(Changeset, Changeset.Schema().load(data=content, many=False, partial=False, unknown="EXCLUDE"))



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import json
import logging
import os
import shutil
import sys
from typing import List, Optional, cast

import aws_cdk.aws_ec2 as ec2
import aws_cdk.aws_efs as efs
import aws_cdk.aws_iam as iam
import aws_cdk.aws_kms as kms
import aws_cdk.aws_s3 as s3
import aws_cdk.aws_ssm as ssm
import jsii
from aws_cdk import core
from aws_cdk.core import App, Aspects, Construct, Duration, Environment, IAspect, IConstruct, Stack, Tags

from aws_orbit.models.changeset import Changeset, load_changeset_from_ssm
from aws_orbit.models.context import Context, ContextSerDe, TeamContext
from aws_orbit.models.manifest import Manifest, ManifestSerDe, TeamManifest
from aws_orbit.remote_files.cdk.team_builders.ec2 import Ec2Builder
from aws_orbit.remote_files.cdk.team_builders.efs import EfsBuilder
from aws_orbit.remote_files.cdk.team_builders.iam import IamBuilder

_logger: logging.Logger = logging.getLogger(__name__)


class Team(Stack):
    def __init__(
        self,
        scope: Construct,
        id: str,
        context: "Context",
        team_name: str,
        team_policies: List[str],
        image: Optional[str],
    ) -> None:
        self.scope = scope
        self.id = id
        self.context: "Context" = context
        self.team_name: str = team_name
        self.team_policies: List[str] = team_policies
        self.image: Optional[str] = image
        super().__init__(
            scope=scope,
            id=id,
            stack_name=id,
            env=Environment(account=self.context.account_id, region=self.context.region),
        )
        Tags.of(scope=cast(IConstruct, self)).add(key="Env", value=f"orbit-{self.context.name}")
        Tags.of(scope=cast(IConstruct, self)).add(key="TeamSpace", value=self.team_name)

        if self.context.networking.vpc_id is None:
            raise ValueError("self.context.networking.vpc_id is None!")
        self.i_vpc = ec2.Vpc.from_vpc_attributes(
            scope=self,
            id="vpc",
            vpc_id=self.context.networking.vpc_id,
            availability_zones=cast(List[str], self.context.networking.availability_zones),
        )
        self.i_isolated_subnets = Ec2Builder.build_subnets(
            scope=self, subnet_manifests=context.networking.isolated_subnets
        )
        self.i_private_subnets = Ec2Builder.build_subnets(
            scope=self, subnet_manifests=context.networking.private_subnets
        )
        administrator_arns: List[str] = []  # A place to add other admins if needed for KMS
        admin_principals = iam.CompositePrincipal(
            *[iam.ArnPrincipal(arn) for arn in administrator_arns],
            iam.ArnPrincipal(f"arn:aws:iam::{self.context.account_id}:root"),
        )
        self.team_kms_key: kms.Key = kms.Key(
            self,
            id="kms-key",
            removal_policy=core.RemovalPolicy.RETAIN,
            enabled=True,
            enable_key_rotation=True,
            policy=iam.PolicyDocument(
                statements=[
                    iam.PolicyStatement(
                        effect=iam.Effect.ALLOW,
                        actions=["kms:*"],
                        resources=["*"],
                        principals=[cast(iam.IPrincipal, admin_principals)],
                    )
                ]
            ),
        )
        self.team_security_group: ec2.SecurityGroup = Ec2Builder.build_team_security_group(
            scope=self, context=context, team_name=self.team_name, vpc=self.i_vpc
        )
        self.policies: List[str] = self.team_policies
        if self.context.scratch_bucket_arn:
            self.scratch_bucket: s3.Bucket = cast(
                s3.Bucket,
                s3.Bucket.from_bucket_attributes(
                    scope=self,
                    id="scratch_bucket",
                    bucket_arn=self.context.scratch_bucket_arn,
                    bucket_name=self.context.scratch_bucket_arn.split(":::")[1],
                ),
            )
        else:
            raise Exception("Scratch bucket was not provided in Manifest ('ScratchBucketArn')")

        self.role_eks_pod = IamBuilder.build_team_role(
            scope=self,
            context=self.context,
            team_name=self.team_name,
            policy_names=self.policies,
            scratch_bucket=cast(s3.IBucket, self.scratch_bucket),
            team_kms_key=self.team_kms_key,
            session_timeout=Duration.hours(12),
        )
        shared_fs_name: str = f"orbit-{context.name}-{self.team_name}-shared-fs"
        if context.shared_efs_fs_id is None:
            raise Exception("Shared EFS File system ID was not provided in Manifest ('SharedEfsFsId')")

        if context.shared_efs_sg_id is None:
            raise Exception("Shared EFS File system security group ID was not provided in Manifest ('SharedEfsSgId')")

        self.shared_fs: efs.FileSystem = cast(
            efs.FileSystem,
            efs.FileSystem.from_file_system_attributes(
                scope=self,
                id=shared_fs_name,
                file_system_id=context.shared_efs_fs_id,
                security_group=ec2.SecurityGroup.from_security_group_id(
                    scope=self, id="team_sec_group", security_group_id=context.shared_efs_sg_id
                ),
            ),
        )

        self.efs_ap: efs.AccessPoint = EfsBuilder.build_file_system_access_point(
            scope=self, team_name=team_name, shared_fs=self.shared_fs, path="shared", ap_name=f"{team_name}-shared"
        )
        self.efs_ap_team_private: efs.AccessPoint = EfsBuilder.build_file_system_access_point(
            scope=self, team_name=team_name, shared_fs=self.shared_fs, path="team", ap_name=f"{team_name}-private"
        )

        team_ssm_parameter_name: str = f"/orbit/{context.name}/teams/{self.team_name}/team"
        self.context_parameter: ssm.StringParameter = ssm.StringParameter(
            scope=self,
            id=team_ssm_parameter_name,
            string_value=json.dumps(
                {
                    "EfsId": self.shared_fs.file_system_id,
                    "EfsApId": self.efs_ap.access_point_id,
                    "EfsApIdPrivate": self.efs_ap_team_private.access_point_id,
                    "EksPodRoleArn": self.role_eks_pod.role_arn,
                    "ScratchBucket": self.scratch_bucket.bucket_name,
                    "TeamKmsKeyArn": self.team_kms_key.key_arn,
                    "TeamSecurityGroupId": self.team_security_group.security_group_id,
                }
            ),
            type=ssm.ParameterType.STRING,
            description="Orbit Workbench Team Context.",
            parameter_name=team_ssm_parameter_name,
            simple_name=False,
            tier=ssm.ParameterTier.INTELLIGENT_TIERING,
        )


def main() -> None:
    _logger.debug("sys.argv: %s", sys.argv)
    if len(sys.argv) == 3:
        context: "Context" = ContextSerDe.load_context_from_ssm(env_name=sys.argv[1], type=Context)
        team_name: str = sys.argv[2]
    else:
        raise ValueError("Unexpected number of values in sys.argv.")

    changeset: Optional["Changeset"] = load_changeset_from_ssm(env_name=context.name)
    _logger.debug("Changeset loaded.")

    team_policies: Optional[List[str]] = None
    image: Optional[str] = None

    if changeset and changeset.teams_changeset and team_name in changeset.teams_changeset.added_teams_names:
        manifest: Optional["Manifest"] = ManifestSerDe.load_manifest_from_ssm(env_name=sys.argv[1], type=Manifest)
        if manifest is None:
            raise ValueError("manifest is None!")
        team_manifest: Optional["TeamManifest"] = manifest.get_team_by_name(name=team_name)
        if team_manifest:
            team_policies = team_manifest.policies
            image = team_manifest.image
        else:
            raise ValueError(f"{team_name} not found in manifest!")
    else:
        team_context: Optional["TeamContext"] = context.get_team_by_name(name=team_name)
        if team_context:
            team_policies = team_context.policies
            image = team_context.image
        else:
            raise ValueError(f"Team {team_name} not found in the context.")

    if team_policies is None:
        raise ValueError("team_policies is None!")

    stack_name: str = f"orbit-{context.name}-{team_name}"
    outdir = os.path.join(".orbit.out", context.name, "cdk", stack_name)
    os.makedirs(outdir, exist_ok=True)
    shutil.rmtree(outdir)
    app = App(outdir=outdir)

    @jsii.implements(core.IAspect)
    class AddDeployPathIAM:
        """ Implementing CDK Aspects to add optional IAM Role prefix to IAM roles """

        def visit(self, node: IConstruct) -> None:
            """ Function to implement a path pattern """

            if isinstance(node, iam.CfnRole):
                node.path = f"/{context.role_prefix}/" if context.role_prefix else "/"

    team_stack = Team(
        scope=app, id=stack_name, context=context, team_name=team_name, team_policies=team_policies, image=image
    )
    Aspects.of(scope=cast(IConstruct, team_stack)).add(cast(IAspect, AddDeployPathIAM()))
    app.synth(force=True)


if __name__ == "__main__":
    main()



#  Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License").
#    You may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import json
import logging
from typing import Any, Dict, List, Optional, cast

from aws_orbit.utils import boto3_client

_logger: logging.Logger = logging.getLogger(__name__)


def get_role(role_name: str) -> Optional[Dict[str, Any]]:
    _logger.debug(f"Getting Role: {role_name}")

    iam_client = boto3_client("iam")
    try:
        return cast(Dict[str, Any], iam_client.get_role(RoleName=role_name))
    except iam_client.exceptions.NoSuchEntityException:
        return None


def get_open_id_connect_provider(account_id: str, open_id_connect_provider_id: str) -> Optional[Dict[str, Any]]:
    open_id_connect_provider_arn = f"arn:aws:iam::{account_id}:oidc-provider/{open_id_connect_provider_id}"
    _logger.debug(f"Getting OpenIDConnectProvider: {open_id_connect_provider_arn}")

    iam_client = boto3_client("iam")
    try:
        return cast(
            Dict[str, Any],
            iam_client.get_open_id_connect_provider(OpenIDConnectProviderArn=open_id_connect_provider_arn),
        )
    except iam_client.exceptions.NoSuchEntityException:
        return None


def update_assume_role_roles(
    account_id: str,
    role_name: str,
    roles_to_add: Optional[List[str]] = None,
    roles_to_remove: Optional[List[str]] = None,
) -> None:
    if not roles_to_add and not roles_to_remove:
        raise Exception("One of roles_to_add or roles_to_remove is required")

    _logger.debug(f"Updating AssumeRolePolicy for {role_name}, Adding: {roles_to_add}, Removing: {roles_to_remove}")

    iam_client = boto3_client("iam")
    assume_role_policy = iam_client.get_role(RoleName=role_name)["Role"]["AssumeRolePolicyDocument"]

    statements = []
    roles_to_add_set = (
        set()
        if roles_to_add is None
        else {f"arn:aws:iam::{account_id}:role/{role}" for role in roles_to_add if get_role(role)}
    )
    roles_to_remove_set = (
        set() if roles_to_remove is None else {f"arn:aws:iam::{account_id}:role/{role}" for role in roles_to_remove}
    )

    _logger.debug("current_policies: %s", assume_role_policy["Statement"])
    for statement in assume_role_policy["Statement"]:
        arn = statement.get("Principal", {}).get("AWS", None)
        if arn in roles_to_remove_set:
            _logger.debug("Removing %s from AssumeRolePolicy", arn)
            continue
        elif arn in roles_to_add_set:
            _logger.debug("AssumeRolePolicy Statement (%s) found containing %s", statement, arn)
            roles_to_add_set.remove(arn)
            statements.append(statement)
        else:
            _logger.debug("Keeping %s in AssumeRolePolicy", statement)
            statements.append(statement)

    for arn in roles_to_add_set:
        _logger.debug("Adding %s to AssumeRolePolicy", arn)
        statements.append({"Effect": "Allow", "Action": "sts:AssumeRole", "Principal": {"AWS": arn}})

    assume_role_policy["Statement"] = statements
    policy_body = json.dumps(assume_role_policy)
    _logger.debug("policy_body: %s", policy_body)
    iam_client.update_assume_role_policy(RoleName=role_name, PolicyDocument=policy_body)


def add_assume_role_statement(role_name: str, statement: Dict[str, Any]) -> None:
    _logger.debug(f"Adding AssumeRolePolicy for {role_name}, Adding: {statement}")

    iam_client = boto3_client("iam")
    assume_role_policy = iam_client.get_role(RoleName=role_name)["Role"]["AssumeRolePolicyDocument"]
    statements = assume_role_policy["Statement"]

    if statement in statements:
        _logger.debug("Skipping Statement already contained by the AssumeRolePolicy")
    else:
        statements.append(statement)
        assume_role_policy["Statement"] = statements
        policy_body = json.dumps(assume_role_policy)
        _logger.debug("policy_body: %s", policy_body)
        iam_client.update_assume_role_policy(RoleName=role_name, PolicyDocument=policy_body)
